#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "microsd.h"

/******************************************************************************/

typedef unsigned char byte;

typedef struct {
  byte code[440];
  byte disk_signature[4];
  byte nulls[2];
  struct {
    byte active;
    byte chs_first[3];
    byte partition_type;
    byte chs_last[3];
    byte lba_first[4];
    byte num_blocks[4];
  } partitions[4];
  byte sig55;
  byte sigaa;
} MBRtype;

typedef struct {
  byte jump[3];
  byte oem[8];
  byte bytes_per_sector[2];
  byte sectors_per_cluster;
  byte reserved_sector_count[2];
  byte num_fats;
  byte max_root_dir_entries[2];
  byte total_sectors_w[2];
  byte media_descriptor;
  byte sectors_per_fat[2];
  byte sectors_per_track[2];
  byte heads[2];
  byte hidden_sectors[4];
  byte total_sectors_l[4];
  byte physical_drive_number;
  byte reserved;
  byte extended_boot_signature;
  byte serial_number[4];
  byte volume_label[11];
  byte fat_system_type[8];
  byte os_code[448];
  byte sig55;
  byte sigaa;
} BootSectorType;

typedef struct {
  byte name[8];
  byte ext[3];
  byte attributes;
  byte reserved;
  byte create_time_10ms;
  byte create_time[2];
  byte create_date[2];
  byte last_access[2];
  byte ea_index[2];
  byte last_mod_time[2];
  byte last_mod_date[2];
  byte first_cluster[2];
  byte file_size[4];
} DirectoryEntryType;

#define c2w(a) ((a)[0] + ((a)[1]<<8))
#define c2l(a) ((a)[0] + ((a)[1]<<8) + ((a)[2]<<16) + ((a)[3]<<24))
#define MIN(a,b) (((a)<(b)) ? (a) : (b))

/******************************************************************************/

static unsigned long part_start, part_size;
static unsigned long fat_start, rootdir_start, data_offset;
static int sectors_per_cluster, sectors_per_rootdir;

static int dcluster, sectors_remaining;
static long cbuf_blockno = 0;
static byte cbuf[512];

unsigned long bytes_remaining;
static byte dbuf[512];
static int dbufptr;

/******************************************************************************/

static void color(int col)
{
  switch (col)
    {
    case -1: printf("\033[0m"); break;
    case 0:
    case 8:
      printf("\033[33m"); break;
    case 4:
    case 12:
      printf("\033[36m"); break;
    }
}
static void
dump_block (byte *buf)
{
  int i, j;
  byte d;
  for (i=0; i<512; i+=16)
    {
      printf("%04x", i);
      printf(" :");
      for (j=0; j<16; j++)
	{
	  d = buf[i+j];
	  color(j);
	  printf(" %02x", d);
	}
      printf(" : ");
      for (j=0; j<15; j++)
	{
	  d = buf[i+j];
	  color(j);
	  putchar (isgraph(d) ? d : '.');
	}
      color(-1);
      printf("\n");
    }
}


/* returns TRUE if we found a usable disk, else FALSE */
static int
scan_disk ()
{
  int i;
  MBRtype *mbr;
  BootSectorType *boots;

  microsd_get_block (0, dbuf);

  mbr = (MBRtype *)dbuf;
  if (mbr->sig55 != 0x55
      || mbr->sigaa != 0xaa)
    {
      printf("AA55 signature wrong: %02x %02x\n", mbr->sigaa, mbr->sig55);
      return 0;
    }

  for (i=0; i<4; i++)
    {
      printf("sda%d : %02x %02x %d %d\n", i,
	     mbr->partitions[i].active,
	     mbr->partitions[i].partition_type,
	     c2l(mbr->partitions[i].lba_first),
	     c2l(mbr->partitions[i].num_blocks));
      if (mbr->partitions[i].partition_type == 6)
	{
	  part_start = c2l(mbr->partitions[i].lba_first);
	  part_size = c2l(mbr->partitions[i].num_blocks);
	  break;
	}
    }
  if (i == 4)
    return 0;

  microsd_get_block (part_start, dbuf);

  boots = (BootSectorType *)dbuf;
  printf("fs \"%8.8s\" sect %d clust %d res %d fats %d\n",
	 boots->oem, c2w(boots->bytes_per_sector), boots->sectors_per_cluster,
	 c2w(boots->reserved_sector_count), boots->num_fats);
  printf("  rootdir %d spfat %d hidden %d total %d %u\n",
	 c2w(boots->max_root_dir_entries), c2w(boots->sectors_per_fat),
	 c2l(boots->hidden_sectors), c2w(boots->total_sectors_w), c2l(boots->total_sectors_l));

  if (c2w(boots->bytes_per_sector) != 512)
    return 0;

  fat_start = part_start + c2w(boots->reserved_sector_count);
  rootdir_start = fat_start + boots->num_fats * c2w(boots->sectors_per_fat);
  sectors_per_rootdir = c2w(boots->max_root_dir_entries) / (512/32);
  data_offset = rootdir_start + sectors_per_rootdir - 2 * boots->sectors_per_cluster;
  sectors_per_cluster = boots->sectors_per_cluster;
  printf("DOS: fat %d root %d data %d spc %d spr %d\n", fat_start, rootdir_start, data_offset,
	 sectors_per_cluster, sectors_per_rootdir);

  return 1;
}

static void
load_cluster_block ()
{
  unsigned long dbuf_blockno;

  if (sectors_remaining == 0)
    {
      unsigned long cblock;

      cblock = fat_start + (dcluster * 2) / 512;
      if (cblock != cbuf_blockno)
	{
	  printf("loading cluster block %d for cluster %d\n", cblock, dcluster);
	  microsd_get_block (cblock, cbuf);
	  cbuf_blockno = cblock;
	}
      dcluster = c2w (cbuf + 2 * (dcluster % (512/2)));
      sectors_remaining = sectors_per_cluster;
    }
  dbuf_blockno = (dcluster * sectors_per_cluster
		  + data_offset
		  + (sectors_per_cluster - sectors_remaining));
  printf("cluster %d sectR %d block %d\n", dcluster, sectors_remaining, dbuf_blockno);
  sectors_remaining --;
  microsd_get_block (dbuf_blockno, dbuf);
  dbufptr = 0;
}

/* Returns TRUE if the file was opened, FALSE otherwise */
static int
open_file (const byte *filename)
{
  int i, j;
  byte match[11];
  DirectoryEntryType *dir = (DirectoryEntryType *)dbuf;

  //printf("%s = ", filename);
  i = 0;
  while (*filename)
    {
      byte c = toupper(*filename++);
      if (c == '.')
	break;
      if (i < 8)
	match[i++] = c;
    }
  while (i < 8)
    match[i++] = ' ';
  while (*filename)
    {
      byte c = toupper(*filename++);
      if (i < 11)
	match[i++] = c;
    }
  while (i < 11)
    match[i++] = ' ';
  //printf("\"%11.11s\"\n", match);

  for (i=0; i<sectors_per_rootdir; i++)
    {
      microsd_get_block (rootdir_start + i, dbuf);
      //dump_block (dbuf);
      for (j=0; j<512/32; j++)
	{
	  if (dir[j].name[0] == 0x00)
	    goto not_found;
	  if (dir[j].name[0] == 0xe5)
	    continue;
	  if (dir[j].attributes == 0x0f)
	    continue;
	  //printf("see   \"%11.11s\"\n", dir[j].name);
	  if (memcmp (match, dir[j].name, 11) == 0)
	    goto found_it;
	}
    }
 not_found:
  return 0;

found_it:
  //printf("found \"%11.11s\"\n", dir[j].name);

  bytes_remaining = c2l(dir[j].file_size);
  dcluster = c2w(dir[j].first_cluster);
  sectors_remaining = sectors_per_cluster;
  load_cluster_block ();
  return 1;
}

/* Returns number of bytes read, 0=EOF */
static int
read_file (void *bufV, int len)
{
  byte *buf = (byte *)bufV;
  int i, rv=0;

  while (len > 0)
    {
      if (bytes_remaining == 0)
	break;

      if (dbufptr < 512)
	{
	  i = MIN (512-dbufptr, len);
	  i = MIN (i, bytes_remaining);
	  memcpy (buf, dbuf+dbufptr, i);
	  buf += i;
	  dbufptr += i;
	  rv += i;
	  len -= i;
	  bytes_remaining -= i;
	  continue;
	}

      load_cluster_block ();
    }
  return rv;
}

main()
{

  while (1)
    {
      printf("waiting for card...\n");
      while (! microsd_card_ready())
	;

      if (scan_disk ())
	if (open_file("spartan.bin"))
	  {
	    int r;
	    char buf[41];
	    while ((r = read_file (buf, 41)) > 0)
	      {
		//printf("\033[31m(r%d)\033[0m", r);
		//fwrite (buf, 1, r, stdout);
	      }
	  }

      printf("waiting for no card...\n");
      while (microsd_card_ready())
	;
      return;
    }
}
