#include "bsp.h"
#include "tty.h"
#include "spartan.h"
#include "fatfs.h"

/* Clock data into fpga on RISING edge.  */

/* Pulse PROG_B 500nS, wait for INIT_B to go high.

   Data changes on falling edge, sampled on rising edge.
   clock HIGH at idle.

   Supply data (or filler) until:

   INIT_B goes low - error
   DONE goes high - success
*/

void
spartan_init ()
{
  register unsigned char new_pd9;

  pd6.b1 = 1; /* CCLK */
  pd6.b2 = 0; /* DOUT - this is output from the fpga, input for us */
  pd6.b3 = 1; /* DIN - this is input to the fpga, output for us */

  pd7.b5 = 1; /* PROG_B */
  pd7.b6 = 0; /* DONE */
  pd7.b7 = 0; /* INIT_B */

  new_pd9 = pd9.b & 0xf0;
  prcr.b = 0x04;
  pd9.b = new_pd9;

  /* We use UART0 in SPI mode */

  u0brg = 0;

  u0mr.smd = 1;		/* Synchronous serial */
  u0mr.ckdir = 0;	/* internal clock */
  u0mr.iopol = 0;	/* not inverted */

  u0c0.clk = 0;		/* 0=f1, 1=f8 */
  u0c0.crd = 1;		/* disable CTS/RTS */
  u0c0.nch = 0;		/* CMOS output */
  u0c0.ckpol = 0;	/* data changes on falling edge */
  u0c0.uform = 1;	/* MSB first */

  u0c1.te = 1;		/* enable transmitter */
  u0c1.re = 1;		/* enable receiver */
  /* If you enable continuous receive, the chip will clock out extra
     bytes just because you're reading the receive buffer.  */
  u0c1.uirrm = 0;	/* continuous receive mode */
  u0c1.uilch = 0;	/* not inverted */

  u0smr.b = 0;		/* no special modes */
  u0smr2.b = 0;		/* no special modes */
  u0smr3.b = 0;		/* clock format */
  u0smr4.b = 0;		/* no I2C */

  s0tic = 0;
  dm1ic = 5;

  ps0.b1 = 1;		/* make P6.1 CLK0 */
  ps0.b3 = 1;		/* make P6.3 TxD0 */
}

void
spartan_diag ()
{
  unsigned char old_done, old_init_b, new_done, new_init_b;

  old_done = F_DONE;
  old_init_b = F_INIT_B;
  F_PROG_B = 0;
  asm("nop");
  asm("nop");
  asm("nop");
  asm("nop");
  asm("nop");
  asm("nop");
  asm("nop");
  asm("nop");
  F_PROG_B = 1;
  wait_ms (1);
  new_done = F_DONE;
  new_init_b = F_INIT_B;
  cprintf("Spartan check: DONE %d INIT_B %d, PROG_B, DONE %d INIT_B %d\n",
	  old_done, old_init_b, new_done, new_init_b);
}

static void
send_byte (unsigned char ch)
{
  int i;
#if 0
  for (i=0; i<8; i++)
    {
      F_CCLK = 0;
      F_DIN = (ch & 0x80) ? 1 : 0;
      asm("nop");
      F_CCLK = 1;
      ch <<= 1;
    }
#else
  while (!u0c1.ti)
    ;
  u0tb = ch;
#endif
}

static volatile int dma1_done = 1;

void __attribute__((interrupt))
dma1_handler()
{
  dma1_done = 1;
}

static void
send_block (unsigned char *ch, int len)
{
  word dmd;

  while (!dma1_done)
    ;

  dmd = get_dmd0();
  dmd &= 0x0f;
  dmd |= 0x90;
  set_dmd0(dmd);	/* single transfer, 8 bits, memory to fixed */
  while (!u0c1.ti)
    ;
  dm1sl = 0x8e;		/* requested, s0tx */
  dma1_done = 0;
  set_dma1(ch+1);
  set_dra1(ch+1);
  set_dsa1(&u0tb);
  set_dct1(len-1);
  set_drc1(len-1);
  u0tb = ch[0];
}

send_block_wait ()
{
  while (!dma1_done)
    ;
}

int
spartan_program ()
{
  char buf[512];
  unsigned int i;
  int r;
  long filesize = 0;
  int end_tics, start_tics;

  if (!open_file ("spartan.bin"))
    {
      cprintf("\033[31mno spartan.bin found\033[0m\n");
      return 0;
    }

  cprintf("\033[32mLoading spartan.bin\033[0m\n");
  F_CCLK = 1;

  F_PROG_B = 0;
  asm("nop");
  asm("nop");
  asm("nop");
  asm("nop");
  asm("nop");
  asm("nop");
  asm("nop");
  asm("nop");
  F_PROG_B = 1;

  while (!F_INIT_B)
    ;

  cprintf("\033[34mbefore programming: done %d init %d\033[0m\n", F_DONE, F_INIT_B);

  start_tics = tics;
  while ((r = read_file (buf, 512)) > 0)
    {
      filesize += r;
#if 0
      for (i=0; i<r; i++)
	send_byte (buf[i]);
#else
      send_block (buf, r);
#endif
      p10.b = p7.b;
    }
  send_block_wait ();
  end_tics = tics;

  cprintf("filling...\n");
  i = 0;
  while (!F_DONE && F_INIT_B && i < 0xffff)
    {
      send_byte (0xff);
      p10.b = p7.b;
      i ++;
    }
  cprintf("\033[34m%d extra bytes, done %d init %d\033[0m\n", i, F_DONE, F_INIT_B);

  cprintf("%l bytes in %d mS\n", filesize, (end_tics - start_tics));

  if (F_DONE)
    return 1;
  else
    return 0;
}
